@@ -1,12 +1,8 @@
 """Implementations related to BASS system."""
 
-import os
-from io import IOBase
-from typing import overload
-
 from requests import Response, session
 
-from . import logger
+from . import is_mock_mode, logger
 
 
 class BassException(Exception):
@@ -68,47 +64,21 @@ class BassProxy:
         self._headers['cookie'] = ';'.join(f'{name}={value}'
                                            for name, value in self._sess.cookies.items())
 
-    @overload
-    def upload_file(self, upload_name: str, fp: str) -> str: ...
-
-    @overload
-    def upload_file(self, upload_name: str, fp: IOBase) -> str: ...
-
-    def upload_file(self, upload_name: str, fp) -> str:
-        """Upload a file to BASS system."""
-        assert self._token, 'must login at first'
-        assert upload_name and isinstance(upload_name, str), f'invalid name: {upload_name}'
-        assert fp, 'nothing to upload'
-        assert isinstance(fp, (str, IOBase)), f'invalid file type: {type(fp)}'
-        if isinstance(fp, str):
-            assert (
-                isinstance(fp, str)
-                and os.path.exists(fp)
-                and os.path.isfile(fp)
-            ), f'{fp} does not exist or is not a file'
-
-        params = {
-            'cmb': 'UploadFile'
-        }
-        headers = self._headers.copy()
-        headers.pop('content-type')
-        if isinstance(fp, str):
-            files = {'File': (upload_name, open(fp, 'rb'))}
-        else:
-            fp.seek(0)
-            files = {'File': (upload_name, fp)}
-        resp = self._sess.post(self._URL, params=params, headers=headers, files=files)
-        resp_data = self._verify_response(resp=resp)
-        file_key = resp_data.get('FileKey')
-        assert file_key and isinstance(file_key, str), f'invalid file key: {file_key=}'
-        return file_key
-
     def launch_task(self, task_id: str, pickle_file_key: str) -> bool:
         assert task_id and isinstance(task_id, str), f'invalid task ID: {task_id}'
         assert (
             pickle_file_key and isinstance(pickle_file_key, str)
         ), f'invalid key of pickle file: {pickle_file_key}'
 
+        if is_mock_mode():
+            return self._launch_task_mock(task_id=task_id, pickle_file_key=pickle_file_key)
+        else:
+            return self._launch_task(task_id=task_id, pickle_file_key=pickle_file_key)
+
+    def _launch_task_mock(self, task_id: str, pickle_file_key: str) -> bool:
+        return True
+
+    def _launch_task(self, task_id: str, pickle_file_key: str) -> bool:
         params = {
             'cmb': 'StartCalculationTaskRunning'
         }
@@ -123,3 +93,36 @@ class BassProxy:
         else:
             logger.error(f'failed to launch task: {resp_data}')
             return False
+
+    def notify_dataset_state(self, task_id: str, verified: bool, cause_of_failuer: str):
+        """Notify varification result of local dataset to the task manager."""
+        assert task_id and isinstance(task_id, str), f'invalid task ID: {task_id}'
+        assert isinstance(verified, bool), f'invalid verification state: {verified}'
+        assert verified or cause_of_failuer, 'Must specify the cause when it is failed.'
+        assert (
+            not cause_of_failuer or isinstance(cause_of_failuer, str)
+        ), f'Invalid cause of failure: {cause_of_failuer}'
+
+        if is_mock_mode():
+            self._notify_dataset_state_mock(task_id=task_id,
+                                            verified=verified,
+                                            cause_of_failuer=cause_of_failuer)
+        else:
+            self._notify_dataset_state(task_id=task_id,
+                                       verified=verified,
+                                       cause_of_failuer=cause_of_failuer)
+
+    def _notify_dataset_state_mock(self, task_id: str, verified: bool, cause_of_failuer: str):
+        pass
+
+    def _notify_dataset_state(self, task_id: str, verified: bool, cause_of_failuer: str):
+        params = {
+            'cmb': 'VerifyTaskDataSetCallback'
+        }
+        data = {
+            'TaskId': task_id,
+            'VerifyResult': verified,
+            'CauseOfFailure': cause_of_failuer
+        }
+        resp = self._sess.post(self._URL, params=params, json=data, headers=self._headers)
+        self._verify_response(resp=resp)
