@@ -3,7 +3,7 @@
 import random
 from typing import List
 from uuid import uuid4
-from .... import DEV_TASK_ID
+from .. import DEV_TASK_ID
 
 
 _INTERSECTION = [
@@ -15,11 +15,6 @@ _INTERSECTION = [
     'fb6eff99-3395-4f07-9c79-e444011c07fb'
 ]
 
-AGGREGATOR_ID = 'QmegaqF4RjBjN3i4NdJy24svtFkJAUXc3YUFqRKqNjaAw4'
-DATA_OWNER_3_ID = 'QmP4DksaGVrFgkNp3d4NxZnnb4RtQucpMqLESJTAEAtuxC'
-DATA_OWNER_4_ID = 'QmWuc1GSaqCkUajoPG5QHfeh72YqU4N7VH7Nx2eP6FyJ9M'
-DATA_OWNER_5_ID = 'QmRj3kV7uQeCndrJvYYFHwfDoQmFdxNkCqpiL7m7VanKfw'
-
 
 def get_task_id() -> str:
     return DEV_TASK_ID
