@@ -1,4 +1,5 @@
 from .psi.rsa_intersecter import *  # noqa
 from .hetero_nn import *  # noqa
 from .hetero_nn_client import HeteroNNCollaboratorScheduler  # noqa
-from ..scheduler import register_metrics  # noqa
+from .secure_hetero_nn import *  # noqa
+from .secure_hetero_nn_client import SecureHeteroNNCollaboratorScheduler  # noqa
