@@ -4,19 +4,23 @@ Reference: https://arxiv.org/pdf/2007.06849.pdf
 """
 
 import io
+import json
 import os
-from tempfile import TemporaryFile
-import time
+import shutil
+import sys
+import traceback
 from abc import ABC, ABCMeta, abstractmethod
+from tempfile import TemporaryFile
 from typing import Dict, List, Set, Tuple
 from zipfile import ZipFile
 
 import torch
 import torch.nn as nn
 import torch.optim as optim
+from torch.utils.tensorboard import SummaryWriter
 
-from .. import logger
-from ..data_channel import GRPCDataChannel
+from .. import get_result_dir, get_runtime_dir, logger
+from ..data_channel import SharedFileDataChannel
 from ..scheduler import ConfigError, Scheduler, TaskFailed
 from .contractor import (CheckinEvent, CollaboratorCompleteEvent,
                          HeteroNNContractor, SendFeatureEvent,
@@ -28,6 +32,10 @@ __all__ = ['HeteroNNHostScheduler']
 _FEATURE_KEY = str
 
 
+class ResetRound(Exception):
+    ...
+
+
 class _SimplifiedOptimizer(ABC):
     """A simplified optimizer tool to facilitate update parameters."""
 
@@ -73,9 +81,12 @@ class HeteroNNScheduler(Scheduler, metaclass=ABCMeta):
         self.id = id
         self.task_id = task_id
         self.is_host = is_initiator
+        self._result_dir = get_result_dir(self.task_id)
+        self._log_dir = os.path.join(self._result_dir, 'tb_logs')
+        self.tb_writer = SummaryWriter(log_dir=self._log_dir)
 
         self.contractor = HeteroNNContractor(task_id=task_id)
-        self.data_channel = GRPCDataChannel(self.contractor)
+        self.data_channel = SharedFileDataChannel(self.contractor)
 
     @abstractmethod
     def load_local_ids(self) -> List[str]:
@@ -156,10 +167,26 @@ class HeteroNNScheduler(Scheduler, metaclass=ABCMeta):
     def _launch_process(self):
         """Run the main process of the task."""
 
-    def _run(self, id: str, task_id: str, is_initiator: bool = False):
+    @abstractmethod
+    def _recover_progress():
+        """Try to recover progress from last running."""
+
+    @abstractmethod
+    def _clean_progress(self):
+        """Clean existing progress data."""
+
+    def _run(self, id: str, task_id: str, is_initiator: bool = False, recover: bool = False):
         self._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
         self.push_log(message='Local context is ready.')
-        self._launch_process()
+        try:
+            if recover:
+                self._recover_progress()
+            else:
+                self._clean_progress()
+            self._launch_process()
+        except Exception:
+            err_stack = ''.join(traceback.format_exception(*sys.exc_info()))
+            self.push_log(err_stack)
 
     def push_log(self, message: str):
         """Push a running log message to the task manager."""
@@ -180,7 +207,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
                  max_rounds: int = 0,
                  calculation_timeout: int = 300,
                  schedule_timeout: int = 30,
-                 data_channel_timeout: Tuple[int, int] = (30, 60),  # TODO 有共享存储后修改
                  log_rounds: int = 0,
                  is_feature_trainable: bool = True) -> None:
         r"""Init.
@@ -199,12 +225,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             :schedule_timeout
                 Seconds to timeout for process scheduling. Takeing off timeout
                 by setting its value to 0.
-            :data_channel_timeout
-                NOTE: This is a temporary configuration. May be removed in next version.
-                Do not design your process depending on it.\n
-                A pair of timeout configuration. The first is seconds to timeout for
-                connecting, and the second is seconds to timeout for transmitting data.
-                Takeing off either one by setting its value to 0.
             :log_rounds
                 The number of rounds to run testing and log the result. Skip it
                 by setting its value to 0.
@@ -219,7 +239,6 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         self.max_rounds = max_rounds
         self.calculation_timeout = calculation_timeout
         self.schedule_timeout = schedule_timeout
-        self.dc_conn_timeout, self.dc_timeout = data_channel_timeout
         self.log_rounds = log_rounds
         self.is_feature_trainable = is_feature_trainable  # TODO 暂时不考虑
 
@@ -228,7 +247,7 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         self.infer_model
         self.infer_optimizer
 
-        self._round = 0
+        self.current_round = 1
         self._partners: List[str] = []
 
         self._example_project_input = None
@@ -400,16 +419,59 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
         super()._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
         if not self.name:
             self.name = f'host_{self.task_id}'
-        self._save_path = os.path.join('models', self.name)
 
         self._init_partners()
+        self._check_in_status = {_partner: False for _partner in self._partners}
+        self._is_gathering_complete = False
+
+        self._runtime_dir = get_runtime_dir(self.task_id)
+        self._context_file = os.path.join(self._runtime_dir, ".context.json")
+        self._checkpoint_dir = os.path.join(self._runtime_dir, 'checkpoint')
+        self._feature_ckpt_file = os.path.join(self._checkpoint_dir, 'feature_model_ckp.pt')
+        self._infer_ckpt_file = os.path.join(self._checkpoint_dir, 'infer_model_ckp.pt')
 
         self.push_log(message='Begin to validate local context.')
         self.validate_context()
 
+    def _recover_progress(self):
+        if not os.path.isfile(self._context_file):
+            raise TaskFailed('Failed to recover progress: missing cached context.')
+
+        with open(self._context_file, 'r') as f:
+            context_info = json.load(f)
+        round = context_info.get('round')
+        feature_ckpt_file = context_info.get('feature_ckpt_file')
+        infer_ckpt_file = context_info.get('infer_ckpt_file')
+        assert round and isinstance(round, int) and round > 0, f'Invalid round: {round} .'
+        assert (
+            feature_ckpt_file and isinstance(feature_ckpt_file, str)
+        ), f'Invalid feature_ckpt_file: {feature_ckpt_file} .'
+        assert (
+            infer_ckpt_file and isinstance(infer_ckpt_file, str)
+        ), f'Invalid infer_ckpt_file: {infer_ckpt_file} .'
+        if not os.path.isfile(feature_ckpt_file) or not os.path.isfile(infer_ckpt_file):
+            raise TaskFailed('Failed to recover progress: missing checkpoint parameters.')
+
+        self.current_round = round
+        with open(feature_ckpt_file, 'rb') as f:
+            state_dict = torch.load(f)
+            self.feature_model.load_state_dict(state_dict)
+        with open(infer_ckpt_file, 'rb') as f:
+            state_dict = torch.load(f)
+            self.infer_model.load_state_dict(state_dict)
+
+    def _clean_progress(self):
+        """Clean existing progress data."""
+        shutil.rmtree(self._runtime_dir, ignore_errors=True)
+        shutil.rmtree(self._result_dir, ignore_errors=True)
+        os.makedirs(self._runtime_dir, exist_ok=True)
+        os.makedirs(self._checkpoint_dir, exist_ok=True)
+        os.makedirs(self._result_dir, exist_ok=True)
+        os.makedirs(self._log_dir, exist_ok=True)
+
     def _is_reach_max_rounds(self) -> bool:
         """Is the max rounds configuration reached."""
-        return self._round >= self.max_rounds
+        return self.current_round >= self.max_rounds
 
     def _validate_feature_dict(self, features: Dict[str, torch.Tensor]):
         """Validate feature format."""
@@ -431,89 +493,83 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             self.push_log(f'Node {self.id} is up.')
 
             self._switch_status(self._GETHORING)
-            self._checkin()
-
-            self._switch_status(self._ID_INTERSECTION)
-            self._make_id_intersection()
+            self._check_in()
 
             self._switch_status(self._READY)
             while self.status == self._READY:
-                self._switch_status(self._SYNCHRONIZING)
-                self._sync_state()
+                try:
+                    self._switch_status(self._SYNCHRONIZING)
+                    self._sync_state()
 
-                self._switch_status(self._IN_A_ROUND)
-                self._run_a_round()
-                self._switch_status(self._READY)
+                    if not self._id_intersection:
+                        self._switch_status(self._ID_INTERSECTION)
+                        self._make_id_intersection()
+
+                    self._switch_status(self._IN_A_ROUND)
+                    self._run_a_round()
+                    self._switch_status(self._READY)
+                except ResetRound:
+                    self.push_log('WARNING: Reset runtime context, there might be an error raised.')
+                    self._switch_status(self._READY)
+                    self._id_intersection = None
+                    continue
 
                 if self.is_task_finished():
                     self.push_log(f'Obtained the final results of task {self.task_id}')
                     self._switch_status(self._FINISHING)
                     self._close_task(is_succ=True)
 
+                self.current_round += 1
+
         except TaskFailed as err:
             logger.exception(err)
             self._close_task(is_succ=False)
 
-    def _checkin(self):
+    def _check_in(self):
         """Check in task and connect every partners."""
         self.push_log('Waiting for participants taking part in ...')
-        check_in_status = dict((_partner, False) for _partner in self._partners)
         for _event in self.contractor.contract_events():
             if isinstance(_event, CheckinEvent):
-                if check_in_status.get(_event.peer_id) is False:
-                    check_in_status[_event.peer_id] = True
-                    self.push_log(f'Welcome a new partner ID: {_event.peer_id}.')
-                    self.push_log(f'There are {sum(check_in_status.values())} partners now.')
-                    self.contractor.respond_checkin(round=self._round,
-                                                    host=self.id,
-                                                    nonce=_event.nonce,
-                                                    requester_id=_event.peer_id)
-                if all(check_in_status.values()):
+                self._handle_check_in(_event)
+                if all(self._check_in_status.values()):
+                    self._is_gathering_complete = True
                     break
-        # it takes seconds to send a response and to consume it in the remote side,
-        # meanwhile, more check in events may arrive soonly. So keep looking for a while.
-        # TODO Ugerly，refactory it!
-        countdown = 10
-        for _event in self.contractor.contract_events(timeout=countdown):
-            if isinstance(_event, CheckinEvent):
-                check_in_status[_event.peer_id] = True
-                self.push_log(f'Welcome a new partner ID: {_event.peer_id}.')
-                self.push_log(f'There are {sum(check_in_status.values())} partners now.')
-                self.contractor.respond_checkin(round=self._round,
-                                                host=self.id,
-                                                nonce=_event.nonce,
-                                                requester_id=_event.peer_id)
         self.push_log('All partners have gethored.')
 
+    def _handle_check_in(self, _event: CheckinEvent):
+        self._check_in_status[_event.peer_id] = True
+        self.push_log(f'Welcome a new partner ID: {_event.peer_id}.')
+        self.push_log(f'There are {sum(self._check_in_status.values())} partners now.')
+        self.contractor.respond_check_in(round=self.current_round,
+                                         host=self.id,
+                                         nonce=_event.nonce,
+                                         requester_id=_event.peer_id)
+        if self._is_gathering_complete:
+            self.contractor.sync_state(round=self.current_round, host=self.id)
+
     def _sync_state(self):
         """Synchronize state before each round, so it's easier to manage the process.
 
         As a host, iterates round, broadcasts and resets context of the new round.
         """
-        self._round += 1
-        self.push_log(f'Initiate state synchronization of round {self._round}.')
-        self.contractor.sync_state(round=self._round, host=self.id)
+        self.push_log(f'Initiate state synchronization of round {self.current_round}.')
+        self.contractor.sync_state(round=self.current_round, host=self.id)
 
-        sync_status = dict((_partner, False) for _partner in self._partners)
+        sync_status = {_partner: False for _partner in self._partners}
         self.push_log('Waiting for synchronization responses ...')
-        for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
+        for _event in self.contractor.contract_events(timeout=0):
             if isinstance(_event, SyncStateResponseEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 if sync_status.get(_event.peer_id) is False:
                     sync_status[_event.peer_id] = True
                     self.push_log(f'Successfully synchronized state with ID: {_event.peer_id}.')
                 if sum(sync_status.values()) == len(self._partners):
-                    return
+                    break
+            elif isinstance(_event, CheckinEvent):
+                self._handle_check_in(_event)
 
-        if sum(sync_status.values()) < len(self._partners):
-            lost_partners = [_peer_id
-                             for _peer_id, status in sync_status.items()
-                             if status is False]
-            self.push_log(f'Task failed because of lost partners: {lost_partners}.')
-            raise TaskFailed(f'Lost partners: {lost_partners}')
-
-        self.push_log(f'Successfully synchronized state in round {self._round}')
+        self.push_log(f'Successfully synchronized state in round {self.current_round}')
 
     def _make_id_intersection(self) -> List[str]:
         """Make PSI and get id intersection for training."""
@@ -523,50 +579,57 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             initiator_id=self.id,
             ids=local_ids,
             collaborator_ids=self._partners,
-            contractor=self.contractor,
-            data_channel_timeout=(self.dc_conn_timeout, self.dc_timeout)
+            contractor=self.contractor
         )
         self._id_intersection = psi_scheduler.make_intersection()
 
     def _run_a_round(self):
-        self._start_round()
-        self.infer_model.train()
-        self.feature_model.train()
-        for _feature_batch, _labels in self.iterate_train_feature(
-            self.feature_model, self.train_ids
-        ):
-            self.push_log('Featured a batch of data.')
-            self._local_features = _feature_batch
-            self._switch_status(self._WAITING_FOR_FEATURES)
-            self._collect_features()
-            self._switch_status(self._GETTING_GRAD)
-            self.train_a_batch(self._feature_fusion_map, _labels)
-            self._switch_status(self._DISTRIBUTING_FEATURE_GRAD)
-            self._distribute_feature_grad()
+        try:
+            self._start_round()
+            self.infer_model.train()
+            self.feature_model.train()
+            for _feature_batch, _labels in self.iterate_train_feature(
+                self.feature_model, self.train_ids
+            ):
+                self.push_log('Featured a batch of data.')
+                self._local_features = _feature_batch
+                self._switch_status(self._WAITING_FOR_FEATURES)
+                self._collect_features()
+                self._switch_status(self._GETTING_GRAD)
+                self.train_a_batch(self._feature_fusion_map, _labels)
+                self._switch_status(self._DISTRIBUTING_FEATURE_GRAD)
+                self._distribute_feature_grad()
 
-        self._switch_status(self._PERSISTING)
-        self._save_model()
-        self._switch_status(self._TESTING)
-        self._check_and_run_test()
-        self._switch_status(self._CLOSING_ROUND)
-        self._close_round()
+            self._switch_status(self._PERSISTING)
+            self._save_model()
+            self._save_runtime_context()
+            self._switch_status(self._TESTING)
+            self._check_and_run_test()
+            self._switch_status(self._CLOSING_ROUND)
+            self._close_round()
+        except TaskFailed as err:
+            err_stack = ''.join(traceback.format_exception(*sys.exc_info()))
+            self.push_log(err_stack)
+            self.contractor.reset_round()
+            raise ResetRound(err)
 
     def _start_round(self):
         """Prepare and start calculation of a round."""
-        self.push_log(f'Begin the training of round {self._round}.')
-        self.contractor.start_round(round=self._round)
-        self.push_log(f'Calculation of round {self._round} is started.')
+        self.push_log(f'Begin the training of round {self.current_round}.')
+        self.contractor.start_round(round=self.current_round)
+        self.push_log(f'Calculation of round {self.current_round} is started.')
 
     def _collect_features(self) -> Dict[str, torch.Tensor]:
         """Collect all input features from all partners."""
         self.push_log('Waiting for collecting all features from partners ...')
-        self.contractor.notify_ready_for_fusion(self._round)
+        self.contractor.notify_ready_for_fusion(self.current_round)
         feature_map: Dict[str, Dict[str, torch.Tensor]] = {
             self.id: {self.feature_key: self._local_features}
         }
         for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, SendFeatureEvent):
-                feature_stream = self.data_channel.receive_stream(_event)
+                feature_stream = self.data_channel.receive_stream(apply_event=_event,
+                                                                  receiver=self.id)
                 buffer = io.BytesIO(feature_stream)
                 features = torch.load(buffer)
                 self._validate_feature_dict(features)
@@ -592,34 +655,42 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
                 tf.seek(0)
                 self.data_channel.send_stream(source=self.id,
                                               target=_partner,
-                                              data_stream=tf.read(),
-                                              connection_timeout=self.dc_conn_timeout,
-                                              timeout=self.dc_timeout)
+                                              data_stream=tf.read())
         self.push_log('Distributed all features grad tensors to collaborators.')
 
     def _save_model(self):
         """Save latest model state."""
-        os.makedirs(self._save_path, exist_ok=True)
-        with open(os.path.join(self._save_path, 'feature_model_ckp.pt'), 'wb') as f:
+        with open(self._feature_ckpt_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
-        with open(os.path.join(self._save_path, 'infer_model_ckp.pt'), 'wb') as f:
+        with open(self._infer_ckpt_file, 'wb') as f:
             torch.save(self.infer_model.state_dict(), f)
         self.push_log('Saved latest parameters locally.')
 
+    def _save_runtime_context(self):
+        """Save runtime context information in case of restoring."""
+        context_info = {
+            'round': self.current_round,
+            'feature_ckpt_file': self._feature_ckpt_file,
+            'infer_ckpt_file': self._infer_ckpt_file
+        }
+        with open(self._context_file, 'w') as f:
+            f.write(json.dumps(context_info, ensure_ascii=False))
+        self.push_log('Saved latest runtime context.')
+
     @torch.no_grad()
     def _check_and_run_test(self):
         """Run test if match configured conditions."""
         if (
-            self._round == 1
-            or (self.log_rounds > 0 and self._round % self.log_rounds == 0)
-            or self._round == self.max_rounds
+            self.current_round == 1
+            or (self.log_rounds > 0 and self.current_round % self.log_rounds == 0)
+            or self.current_round == self.max_rounds
         ):
             self.push_log('Start a round of test.')
 
             self.feature_model.eval()
             self.infer_model.eval()
 
-            self.contractor.start_test_round(round=self._round)
+            self.contractor.start_test_round(round=self.current_round)
 
             batched_host_features = []
             batched_labels = []
@@ -642,15 +713,16 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             self.push_log('Complete a round of test.')
 
         self.push_log('Skip or close a round of testing.')
-        self.contractor.close_test_round(round=self._round)
+        self.contractor.close_test_round(round=self.current_round)
 
     def _wait_for_testing_features(self):
         """Wait for collecting test dataset features."""
         self.push_log('Waiting for collecting test dataset features ...')
         self._batched_test_features = []
-        for _event in self.contractor.contract_events():
+        for _event in self.contractor.contract_events(timeout=self.calculation_timeout):
             if isinstance(_event, SendTestFeatureEvent):
-                stream = self.data_channel.receive_stream(_event)
+                stream = self.data_channel.receive_stream(apply_event=_event,
+                                                          receiver=self.id)
                 buffer = io.BytesIO(stream)
                 batched_features: dict = torch.load(buffer)
                 _key, _feature_list = batched_features.copy().popitem()
@@ -662,8 +734,8 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
 
     def _close_round(self):
         """Close current round when finished."""
-        self.contractor.close_round(round=self._round)
-        self.push_log(f'The training of Round {self._round} complete.')
+        self.contractor.close_round(round=self.current_round)
+        self.push_log(f'The training of Round {self.current_round} complete.')
 
     def _close_task(self, is_succ: bool = True):
         """Close the task.
@@ -693,29 +765,37 @@ class HeteroNNHostScheduler(HeteroNNScheduler):
             Local paths of the report file and model file.
         """
         self.push_log('Generating task achievement files ...')
+
+        # TODO remove metrics_files later
         metrics_files = []
         for _name, _metrics in self._metrics_bucket.items():
-            _file = f'{os.path.join(self._save_path, _name)}.csv'
+            _file = f'{os.path.join(self._result_dir, _name)}.csv'
             _metrics.to_csv(_file)
             metrics_files.append(_file)
-        report_file = os.path.join(self._save_path, 'report.zip')
+        report_file = os.path.join(self._result_dir, 'report.zip')
         with ZipFile(report_file, 'w') as report_zip:
             for _file in metrics_files:
-                report_zip.write(_file)
+                report_zip.write(_file, os.path.basename(_file))
+            for path, _, filenames in os.walk(self._log_dir):
+                rel_dir = os.path.relpath(path=path, start=self._result_dir)
+                rel_dir = rel_dir.lstrip('.')  # ./file => file
+                for _file in filenames:
+                    rel_path = os.path.join(rel_dir, _file)
+                    report_zip.write(os.path.join(path, _file), rel_path)
         report_file_path = os.path.abspath(report_file)
 
         # torch.jit doesn't work with a TemporaryFile
-        feature_model_file = os.path.join(self._save_path,
+        feature_model_file = os.path.join(self._result_dir,
                                           f'feature_model_{self.feature_key}.pt')
         with open(feature_model_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
-        infer_model_file = f'{os.path.join(self._save_path, "infer_model.pt")}'
+        infer_model_file = f'{os.path.join(self._result_dir, "infer_model.pt")}'
         with open(infer_model_file, 'wb') as f:
             torch.save(self.infer_model.state_dict(), f)
-        model_file = os.path.join(self._save_path, 'model.zip')
+        model_file = os.path.join(self._result_dir, 'model.zip')
         with ZipFile(model_file, 'w') as model_zip:
-            model_zip.write(feature_model_file)
-            model_zip.write(infer_model_file)
+            model_zip.write(feature_model_file, os.path.basename(feature_model_file))
+            model_zip.write(infer_model_file, os.path.basename(infer_model_file))
         model_file_path = os.path.abspath(model_file)
 
         self.push_log('Task achievement files are ready.')
