@@ -11,7 +11,7 @@ from Crypto.PublicKey import RSA
 from Cryptodome import Random
 
 from ... import logger
-from ...data_channel import GRPCDataChannel
+from ...data_channel import SharedFileDataChannel
 from . import PSICollaboratorScheduler, PSIInitiatorScheduler
 from .rsa_psi_contractor import (DistributingIntersectionEvent, PSIInitEvent,
                                  PublicKeyEvent, ReadyForSignedIdsEvent,
@@ -177,14 +177,15 @@ class RSAPSIInitiatorScheduler(PSIInitiatorScheduler):
                  task_id: str,
                  initiator_id: str,
                  ids: List[str],
-                 collaborator_ids: List[str]) -> None:
+                 collaborator_ids: List[str],
+                 contractor: RSAContractor = None) -> None:
         super().__init__(task_id=task_id,
                          initiator_id=initiator_id,
                          ids=ids,
                          collaborator_ids=collaborator_ids)
         self._switch_status(self._INIT)
-        self._contractor = RSAContractor(task_id=task_id)
-        self._data_channel = GRPCDataChannel(contractor=self._contractor)
+        self._contractor = contractor or RSAContractor(task_id=task_id)
+        self._data_channel = SharedFileDataChannel(contractor=self._contractor)
 
         self.coll_init_map: Dict[str, _PSIInitiator] = {}
         self.coll_signed_ids_map: Dict[str, _SignedIdsInfo] = {}
@@ -242,7 +243,8 @@ class RSAPSIInitiatorScheduler(PSIInitiatorScheduler):
         self._contractor.notify_ready_for_signed_ids(initiator=self.initiator_id)
         for _event in self._contractor.contract_events():
             if isinstance(_event, SendingSignedIdsEvent):
-                signed_ids_stream = self._data_channel.receive_stream(_event)
+                signed_ids_stream = self._data_channel.receive_stream(
+                    apply_event=_event, receiver=self.initiator_id)
                 signed_ids = _SignedIdsInfo.from_bytes(signed_ids_stream)
                 self.coll_signed_ids_map[_event.source] = signed_ids
             if len(self.coll_signed_ids_map) == len(self.collaborator_ids):
@@ -280,13 +282,17 @@ class RSAPSICollaboratorScheduler(PSICollaboratorScheduler):
     _SENDING_SIGNED_IDS = 'sending_signed_ids'
     _RECEIVING_INTERSECTION = 'receiving_intersection'
 
-    def __init__(self, task_id: str, collaborator_id: str, ids: List[str]) -> None:
+    def __init__(self,
+                 task_id: str,
+                 collaborator_id: str,
+                 ids: List[str],
+                 contractor: RSAContractor = None) -> None:
         super().__init__(task_id=task_id,
                          collaborator_id=collaborator_id,
                          ids=ids)
         self._switch_status(self._INIT)
-        self._contractor = RSAContractor(task_id=task_id)
-        self._data_channel = GRPCDataChannel(contractor=self._contractor)
+        self._contractor = contractor or RSAContractor(task_id=task_id)
+        self._data_channel = SharedFileDataChannel(contractor=self._contractor)
         self._coll = None
 
         self.intersection: Set[str] = {}
@@ -325,7 +331,9 @@ class RSAPSICollaboratorScheduler(PSICollaboratorScheduler):
         """Wait for blind ids to sign."""
         for _event in self._contractor.contract_events():
             if isinstance(_event, SendingBlindIdsEvent):
-                data_stream = self._data_channel.receive_stream(_event)
+                data_stream = self._data_channel.receive_stream(
+                    apply_event=_event, receiver=self.collaborator_id
+                )
                 blind_ids = json.loads(data_stream.decode(encoding=_ENCODING))
                 return blind_ids
 
@@ -346,7 +354,9 @@ class RSAPSICollaboratorScheduler(PSICollaboratorScheduler):
         """Wait for receiving intersection result."""
         for _event in self._contractor.contract_events():
             if isinstance(_event, DistributingIntersectionEvent):
-                intersection_stream = self._data_channel.receive_stream(_event)
+                intersection_stream = self._data_channel.receive_stream(
+                    apply_event=_event, receiver=self.collaborator_id
+                )
                 _intersection = json.loads(intersection_stream.decode(encoding=_ENCODING))
                 self.intersection = set(self._coll.signed_to_raw(_signed_id)
                                         for _signed_id in _intersection)
