@@ -1,5 +1,7 @@
+import json
 import os
 import random
+import shutil
 from abc import abstractmethod
 from tempfile import TemporaryFile
 from typing import List, Tuple
@@ -9,14 +11,15 @@ import torch
 import torch.nn as nn
 import torch.optim as optim
 
-from .. import logger
+from .. import get_runtime_dir, logger
 from ..scheduler import ConfigError, TaskComplete, TaskFailed
+from .hetero_nn import ResetRound
 from .psi import RSAPSICollaboratorScheduler
 from .secure_contractor import (CheckinResponseEvent, CloseRoundEvent,
                                 CloseTestRoundEvent, CompleteTaskEvent,
                                 FailTaskEvent, ReadyForFusionEvent,
                                 ReadyForNoisedProjectionEvent,
-                                ReadyForNoisedWGradEvent,
+                                ReadyForNoisedWGradEvent, ResetRoundEvent,
                                 SendCipherFeatureGradEvent,
                                 SendCipherPorjectionEvent,
                                 SendCipherWGradEvent, StartRoundEvent,
@@ -39,7 +42,6 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                  project_layer_lr: int,
                  name: str = None,
                  schedule_timeout: int = 30,
-                 data_channel_timeout: Tuple[int, int] = (30, 60),  # TODO 有共享存储后修改
                  is_feature_trainable: bool = True) -> None:
         """Init.
 
@@ -54,12 +56,6 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
             :schedule_timeout
                 Seconds to timeout for process scheduling. Takeing off timeout
                 by setting its value to 0.
-            :data_channel_timeout
-                NOTE: This is a temporary configuration. May be removed in next version.
-                Do not design your process depending on it.
-                A pair of timeout configuration. The first is seconds to timeout for
-                connecting, and the second is seconds to timeout for transmitting data.
-                Takeing off either one by setting its value to 0.
             :is_feature_trainable
                 Decide whether or not train the feature model
         """
@@ -70,12 +66,11 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self.project_layer_lr = project_layer_lr
         self.name = name
         self.schedule_timeout = schedule_timeout
-        self.dc_conn_timeout, self.dc_timeout = data_channel_timeout
         self.is_feature_trainable = is_feature_trainable
 
         self._validate_config()
 
-        self._round = 0
+        self.current_round = 0
 
         self.host = None
 
@@ -144,7 +139,6 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         super()._setup_context(id=id, task_id=task_id, is_initiator=is_initiator)
         if not self.name:
             self.name = f'guest_{self.task_id}'
-        self._save_path = os.path.join('models', self.name)
 
         self._he_context = ts.context(ts.SCHEME_TYPE.CKKS,
                                       poly_modulus_degree=2**12,
@@ -153,28 +147,66 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self._he_context_serialize = self._he_context.serialize()
         self._epsilon_acc = random.gauss(0, 0.1)
 
+        self._runtime_dir = get_runtime_dir(self.task_id)
+        self._context_file = os.path.join(self._runtime_dir, ".context.json")
+        self._checkpoint_dir = os.path.join(self._runtime_dir, 'checkpoint')
+        self._feature_ckpt_file = os.path.join(self._checkpoint_dir, "feature_model_ckp.pt")
+
         self.push_log(message='Begin to validate local context.')
         self.validate_context()
 
+    def _recover_progress(self):
+        if not os.path.isfile(self._context_file):
+            raise TaskFailed('Failed to recover progress: missing cached context.')
+
+        with open(self._context_file, 'r') as f:
+            context_info = json.load(f)
+        feature_ckpt_file = context_info.get('feature_ckpt_file')
+        assert (
+            feature_ckpt_file and isinstance(feature_ckpt_file, str)
+        ), f'Invalid feature_ckpt_file: {feature_ckpt_file} .'
+        if not os.path.isfile(feature_ckpt_file):
+            raise TaskFailed('Failed to recover progress: missing checkpoint parameters.')
+
+        self.current_round = round
+        with open(feature_ckpt_file, 'rb') as f:
+            state_dict = torch.load(f)
+            self.feature_model.load_state_dict(state_dict)
+
+    def _clean_progress(self):
+        """Clean existing progress data."""
+        shutil.rmtree(self._runtime_dir, ignore_errors=True)
+        shutil.rmtree(self._result_dir, ignore_errors=True)
+        os.makedirs(self._runtime_dir, exist_ok=True)
+        os.makedirs(self._checkpoint_dir, exist_ok=True)
+        os.makedirs(self._result_dir, exist_ok=True)
+
     def _launch_process(self):
         try:
             assert self.status == self._INIT, 'must begin from initial status'
             self.push_log(f'Node {self.id} is up.')
 
             self._switch_status(self._GETHORING)
-            self._checkin()
-
-            self._switch_status(self._ID_INTERSECTION)
-            self._make_id_intersection()
+            self._check_in()
 
             self._switch_status(self._READY)
             while self.status == self._READY:
-                self._switch_status(self._SYNCHRONIZING)
-                self._sync_state()
+                try:
+                    self._switch_status(self._SYNCHRONIZING)
+                    self._sync_state()
 
-                self._switch_status(self._IN_A_ROUND)
-                self._run_a_round()
-                self._switch_status(self._READY)
+                    if not self._id_intersection:
+                        self._switch_status(self._ID_INTERSECTION)
+                        self._make_id_intersection()
+
+                    self._switch_status(self._IN_A_ROUND)
+                    self._run_a_round()
+                    self._switch_status(self._READY)
+                except ResetRound:
+                    self.push_log('WARNING: Reset runtime context, there might be an error raised.')
+                    self._switch_status(self._READY)
+                    self._id_intersection = None
+                    continue
 
         except TaskComplete:
             logger.info('Task complete')
@@ -184,7 +216,7 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
             logger.exception(err)
             self._close_task(is_succ=False)
 
-    def _checkin(self):
+    def _check_in(self):
         """Check in task."""
         is_checked_in = False
         # the host may be in special state so can not response
@@ -192,12 +224,12 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self.push_log('Checking in the task ...')
         while not is_checked_in:
             nonce = self.contractor.checkin(peer_id=self.id)
-            logger.debug('_wait_for_checkin_response ...')
+            logger.debug('_wait_for_check_in_response ...')
             for _event in self.contractor.contract_events(timeout=self.schedule_timeout):
                 if isinstance(_event, CheckinResponseEvent):
                     if _event.nonce != nonce:
                         continue
-                    self._round = _event.round
+                    self.current_round = _event.round
                     self.host = _event.host
                     is_checked_in = True
                     break
@@ -215,8 +247,8 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self.push_log('Waiting for synchronizing state with the host ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, SyncStateEvent):
-                self._round = _event.round
-                self.contractor.respond_sync_state(round=self._round,
+                self.current_round = _event.round
+                self.contractor.respond_sync_state(round=self.current_round,
                                                    peer_id=self.id,
                                                    host=_event.host)
                 self.push_log('Successfully synchronized state with the host.')
@@ -225,8 +257,10 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                 raise TaskFailed('Aborted by host.')
             elif isinstance(_event, CompleteTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
-        self.push_log(f'Successfully synchronized state in round {self._round}')
+        self.push_log(f'Successfully synchronized state in round {self.current_round}')
 
     def _make_id_intersection(self) -> List[str]:
         """Make PSI and get id intersection for training."""
@@ -235,8 +269,7 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
             task_id=self.task_id,
             collaborator_id=self.id,
             ids=local_ids,
-            contractor=self.contractor,
-            data_channel_timeout=(self.dc_conn_timeout, self.dc_timeout)
+            contractor=self.contractor
         )
         self._id_intersection = psi_scheduler.collaborate_intersection()
 
@@ -266,6 +299,7 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
 
         self._switch_status(self._PERSISTING)
         self._save_model()
+        self._save_runtime_context()
 
         self._switch_status(self._TESTING)
         self._wait_for_testing_round()
@@ -273,20 +307,22 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self._switch_status(self._CLOSING_ROUND)
         self._wait_for_closing_round()
 
-        self.push_log(f'ID: {self.id} finished training task of round {self._round}.')
+        self.push_log(f'ID: {self.id} finished training task of round {self.current_round}.')
 
     def _wait_for_starting_round(self):
         """Wait for starting a new round of training ..."""
-        self.push_log(f'Waiting for training of round {self._round} begin ...')
+        self.push_log(f'Waiting for training of round {self.current_round} begin ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, StartRoundEvent):
                 assert (
-                    _event.round == self._round
-                ), f'Lost synchronization, host: {_event.round}; local: {self._round}.'
-                self.push_log(f'Training of round {self._round} begins.')
+                    _event.round == self.current_round
+                ), f'Lost synchronization, host: {_event.round}; local: {self.current_round}.'
+                self.push_log(f'Training of round {self.current_round} begins.')
                 return
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _send_feature_cipher(self):
         """Send local features of a batch of data to the host."""
@@ -294,11 +330,13 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         for _event in self.contractor.contract_events():
             if isinstance(_event, ReadyForFusionEvent):
                 assert (
-                    _event.round == self._round
-                ), f'Lost synchronization, host: {_event.round}; local: {self._round}.'
+                    _event.round == self.current_round
+                ), f'Lost synchronization, host: {_event.round}; local: {self.current_round}.'
                 break
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
         self.push_log('Begin to send features.')
         cipher_tensor = ts.ckks_tensor(self._he_context, self._local_features.detach())
@@ -307,9 +345,7 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                                             context=self._he_context_serialize)
         self.data_channel.send_stream(source=self.id,
                                       target=self.host,
-                                      data_stream=cipher_feature.to_bytes(),
-                                      connection_timeout=self.dc_conn_timeout,
-                                      timeout=self.dc_timeout)
+                                      data_stream=cipher_feature.to_bytes())
         self.push_log('Sending features complete.')
 
     def _wait_for_cipher_projection(self):
@@ -317,7 +353,8 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self.push_log('Waiting for receiving cipher projection ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, SendCipherPorjectionEvent):
-                stream = self.data_channel.receive_stream(_event)
+                stream = self.data_channel.receive_stream(apply_event=_event,
+                                                          receiver=self.id)
                 cipher_projection = _CKKSTensorWrapper.from_bytes(stream)
                 cipher_tensor = ts.ckks_tensor_from(context=self._he_context,
                                                     data=cipher_projection.cipher)
@@ -328,6 +365,8 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                 return
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _send_noised_projection(self):
         """Send noised projection to the host for fusion."""
@@ -339,19 +378,20 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                     tf.seek(0)
                     self.data_channel.send_stream(source=self.id,
                                                   target=self.host,
-                                                  data_stream=tf.read(),
-                                                  connection_timeout=self.dc_conn_timeout,
-                                                  timeout=self.dc_timeout)
+                                                  data_stream=tf.read())
                 return
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _wait_for_w_grad(self):
         """Wait for cipher grad of project layer W to decrypt."""
         self.push_log('Waiting for cipher grad of project layer W ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, SendCipherWGradEvent):
-                stream = self.data_channel.receive_stream(_event)
+                stream = self.data_channel.receive_stream(apply_event=_event,
+                                                          receiver=self.id)
                 cipher_grad = ts.ckks_tensor_from(self._he_context, stream)
                 # sample from (-0.005, 0.015) to decrease the overrall speed of
                 # accumulation, thus facilitate convergence
@@ -362,6 +402,8 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                 break
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
         self.push_log('Received and decrypted cipher grad of project layer W.')
 
     def _send_noised_w_grad(self):
@@ -376,10 +418,10 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                                                     context=self._he_context_serialize)
                 self.data_channel.send_stream(source=self.id,
                                               target=self.host,
-                                              data_stream=noised_w_grad.to_bytes(),
-                                              connection_timeout=self.dc_conn_timeout,
-                                              timeout=self.dc_timeout)
+                                              data_stream=noised_w_grad.to_bytes())
                 break
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
         self.push_log('Sent noised grad of project layer W to host.')
 
     def _wait_for_feature_grad(self):
@@ -387,27 +429,38 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
         self.push_log('Waiting for cipher grad of feature model output ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, SendCipherFeatureGradEvent):
-                stream = self.data_channel.receive_stream(_event)
+                stream = self.data_channel.receive_stream(apply_event=_event,
+                                                          receiver=self.id)
                 cipher_grad = ts.ckks_tensor_from(self._he_context, stream)
                 self._feature_grad = torch.tensor(cipher_grad.decrypt().tolist())
                 break
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
         self.push_log('Received and decrypted cipher grad of feature model output.')
 
     def _save_model(self):
         """Save latest model state."""
-        os.makedirs(self._save_path, exist_ok=True)
-        with open(f'{os.path.join(self._save_path, "feature_model_ckp.pt")}', 'wb') as f:
+        with open(self._feature_ckpt_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
         self.push_log('Saved latest parameters locally.')
 
+    def _save_runtime_context(self):
+        """Save runtime context information in case of restoring."""
+        context_info = {
+            'feature_ckpt_file': self._feature_ckpt_file
+        }
+        with open(self._context_file, 'w') as f:
+            f.write(json.dumps(context_info, ensure_ascii=False))
+        self.push_log('Saved latest runtime context.')
+
     def _wait_for_testing_round(self):
         """Wait for handle a round of testing."""
         self.push_log('Waiting for start a round of testing ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, StartTestRoundEvent):
                 assert (
-                    _event.round == self._round
-                ), f'Lost synchronization, host: {_event.round}; local: {self._round}.'
+                    _event.round == self.current_round
+                ), f'Lost synchronization, host: {_event.round}; local: {self.current_round}.'
 
                 self.feature_model.eval()
                 for _batch_features in self.iterate_test_feature(self.feature_model,
@@ -426,17 +479,21 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
                 return
             elif isinstance(_event, FailTaskEvent):
                 raise TaskFailed('Aborted by host.')
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _wait_for_closing_round(self):
         """Wait for closing current round of training."""
-        self.push_log(f'Waiting for closing signal of training round {self._round} ...')
+        self.push_log(f'Waiting for closing signal of training round {self.current_round} ...')
         for _event in self.contractor.contract_events():
             if isinstance(_event, CloseRoundEvent):
-                if _event.round != self._round:
+                if _event.round != self.current_round:
                     continue
                 return
             elif isinstance(_event, CompleteTaskEvent):
                 raise TaskComplete()
+            elif isinstance(_event, ResetRoundEvent):
+                raise ResetRound()
 
     def _close_task(self, is_succ: bool = True):
         """Close the task and upload the final parameters."""
@@ -459,8 +516,9 @@ class SecureHeteroNNCollaboratorScheduler(SecureHeteroNNScheduler):
             Local paths of the model file.
         """
         self.push_log('Generating task achievement files ...')
+
         # torch.jit doesn't work with a TemporaryFile
-        feature_model_file = os.path.join(self._save_path,
+        feature_model_file = os.path.join(self._result_dir,
                                           f'feature_model_{self.feature_key}.pt')
         with open(feature_model_file, 'wb') as f:
             torch.save(self.feature_model.state_dict(), f)
